import datetime
import isoweek
import datetime

''' weeks start on mondays '''
def first_date_of_week(any_datetime):
  year, iso_week, iso_weekday = any_datetime.isocalendar()
  return any_datetime - datetime.timedelta(days=(iso_weekday-1))

def last_date_of_week(any_datetime):
  year, iso_week, iso_weekday = any_datetime.isocalendar()
  return first_date_of_week(any_datetime) + datetime.timedelta(days=6) # last date is 6 days after the first date

def is_leap_day(date):
  return date.month == 2 and date.day == 29

def is_in_leap_week(date):
  return isoweek.Week.withdate(date).week == 53

def is_in_week_prior_leap_week(date):
  same_day_next_week = date + datetime.timedelta(days=7)
  return is_in_leap_week(same_day_next_week)

def has_leap_week(year: int):
  return isoweek.Week(year, 53).week == 53

def last_year(date):
  return isoweek.Week.withdate(date).year - 1

def days_since_same_date_last_year(date):
  return 7 * 53 if is_in_leap_week(date) or has_leap_week(last_year(date)) else 7 * 52

def comp_yearwise(datetime_or_date, n_years):
    dt : datetime.datetime = safe_datetime(datetime_or_date)
    comp_days = 0
    dt_week = isoweek.Week.withdate(dt)
    while n_years >= 1:
        if n_years==1 and comp_days == 0:
            weeks = 52
        else:
            weeks = isoweek.Week.last_week_of_year(dt_week.year - n_years).week
        comp_days += weeks*7
        n_years = n_years-1
    comp_date = dt - datetime.timedelta(days=comp_days)
    return comp_date

def days_between(datetime_or_date_a, datetime_or_date_b):

    dt_a : datetime.datetime = safe_datetime(datetime_or_date_a)
    dt_b : datetime.datetime = safe_datetime(datetime_or_date_b)

    delta = dt_a - dt_b
    
    return delta.days

def safe_datetime(datetime_or_date):
    if type(datetime_or_date) == datetime.date:
        return datetime.datetime.combine(datetime_or_date, datetime.datetime.min.time())
    elif type(datetime_or_date) == datetime.datetime:
        return datetime_or_date
    else:
        raise TypeError


import openpyxl as xl
import pandas as pd
from deprecated import deprecated

def workbook_to_bytes(wb: xl.workbook.workbook.Workbook):
    for x in wb.worksheets:
        x.protection.enable() # always do this before exporting because they automatically become unprotected somehow

    return xl.writer.excel.save_virtual_workbook(wb)

def df_to_worksheet(df: pd.DataFrame, cell: xl.cell.cell.Cell):
    ws = cell.parent

    for r in range(df.shape[0]):
        for c in range(df.shape[1]):
            cell.offset(r+1, c).value = df.iloc[r, c]

@deprecated(version='0.3.0', reason='renamed to df_to_worksheet' )
def df_to_sh(df: pd.DataFrame, cell: xl.cell.cell.Cell):
    df_to_worksheet(df, cell)



import datetime, calendar, os

def readable_date_min(date):
    return date.strftime("%m/%d")

def readable_date(date):
    return date.strftime("%m/%d/%Y")

def readable_week_min(date):
    return f'Week {date.isocalendar()[1]}'

def readable_week(date):
    return f'{date.isocalendar()[0]} W{date.isocalendar()[1]}'    

def readable_week_and_date(date):
    date_str = readable_date_min(date)
    week_str = readable_week(date)

    return f'{week_str} - {date_str}'

def filename_date_min(date):
    return date.strftime("%m.%d")

def filename_date(date):
    return date.strftime("%Y.%m.%d")

def filename_week_min(date):
    return readable_week_min(date)

def filename_week(date):
    return readable_week(date)

def filename_week_and_date(date):
    date_str = filename_date_min(date)
    week_str = filename_week(date)
    return f'{week_str} - {date_str}'



from typing import Dict
import datetime
import argparse
from functools import wraps
import xml.etree.cElementTree as ET
from io import StringIO
import time
import pendulum
import warnings

def parse_templated(templated: Dict, extra_keywords: Dict=None, map_value=lambda x:x, map_key=lambda x:x):
    assert 'templates' in templated
    assert 'keywords' in templated or 'template_dict' in templated

    if 'template_dict' in templated:
        warnings.warn(
            "`template_dict` key will be renamed in version 3, use `keywords` instead",
            PendingDeprecationWarning
        )

    keywords_compat = templated.get("keywords", templated.get("template_dict")) # fallback to old name for compat
    keywords = {**keywords_compat, **extra_keywords} if extra_keywords else keywords_compat

    # TODO: implement support for deep recursive formatted keywords
    formatted_keywords = {
        key: value.format(**keywords)
        for key, value in keywords.items()
    }

    return {
        map_key(key): map_value(value.format(**formatted_keywords))
        for key, value in templated["templates"].items()
    }

def pen2dt(pen: pendulum.DateTime) -> datetime.datetime:
    return datetime.datetime.fromtimestamp(pen.timestamp())

def parse_xml(xml_str):
    # Remove invalid characters in the XML
    xml_str_sanitized = xml_str.replace('&#x1A','').replace('&#x00','').replace('&#x01','')

    # convert the data from xml to an array before returning
    data = []

    for event, elem in ET.iterparse(StringIO(xml_str_sanitized)):
        if elem.tag == "row":
            data.append(elem.attrib)
            elem.clear()

    return data

def ymd_arg(date_str):
    '''parses 'yyyy-mm-dd' formatted date str into a python date object'''
    if date_str is None:
        return date_str
    else:
        try:
            return datetime.datetime.strptime(date_str, '%Y-%m-%d').date()
        except ValueError:
            msg = "Invalid date argument '{}'".format(date_str)
            raise argparse.ArgumentTypeError(msg)


def timeit(func):
    @wraps(func)
    def _time_it(*args, **kwargs):
        start = int(round(time.time() * 1000))
        try:
            return func(*args, **kwargs)
        finally:
            end_ = int(round(time.time() * 1000)) - start
            print(f"⏱ timeit: {func.__module__}.{func.__name__} - [{end_ if end_ > 0 else 0} ms]")
    return _time_it

def sql_arr_lit(lst: list) -> str : 
    assert len(lst) > 0, 'non-empty list required to make sql array literal'
    return f"({repr(lst[0])})" if len(lst) == 1 else repr(tuple(lst))